from highton.call_mixins import Call


class UpdateCallMixin(Call):
    """
    A mixin to update a singe object of an highrise endpoint

    """

    def update(self):
        """
        Updates the object

        :return:
        :rtype: response
        """
        return self._put_request(
            data=self.element_to_string(
                self.encode()
            ),
            endpoint=self.ENDPOINT + '/' + str(self.id)
        )



from highton import fields
from highton.highton_constants import HightonConstants
from highton.models import HightonModel


class Category(
    HightonModel,
):
    """

    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar name: fields.StringField(name=HightonConstants.NAME)
    """
    TAG_NAME = HightonConstants.CATEGORY

    def __init__(self, **kwargs):
        self.name = fields.StringField(name=HightonConstants.NAME)

        super().__init__(**kwargs)



from highton.call_mixins import Call
from highton import fields


class ListNoteCallMixin(Call):
    """
    A mixin to get all notes of inherited class
    These could be: people || companies || kases || deals

    """

    NOTES_OFFSET = 25

    def list_notes(self, page=0, since=None):
        """
        Get the notes of current object

        :param page: the page starting at 0
        :type since: int
        :param since: get all notes since a datetime
        :type since: datetime.datetime
        :return: the notes
        :rtype: list
        """
        from highton.models.note import Note
        params = {'n': int(page) * self.NOTES_OFFSET}
        if since:
            params['since'] = since.strftime(self.COLLECTION_DATETIME)

        return fields.ListField(
            name=self.ENDPOINT,
            init_class=Note
        ).decode(
            self.element_from_string(
                self._get_request(
                    endpoint=self.ENDPOINT + '/' + str(self.id) + '/' + Note.ENDPOINT,
                    params=params
                ).text
            )
        )



from abc import ABCMeta

from highton import fields
from highton.highton_constants import HightonConstants
from highton.parsing.xml_decoder import XMLDecoder
from highton.parsing.xml_encoder import XMLEncoder


class HightonModel(XMLDecoder, XMLEncoder, metaclass=ABCMeta):
    """
    This class inherit from XMLDecoder and XMLEncoder which allows every HightonModel to be encoded and decoded in xml.

    :ivar id: fields.IntegerField(name=HightonConstants.ID)

    """
    ENDPOINT = None

    def __init__(self, **kwargs):
        self.id = fields.IntegerField(name=HightonConstants.ID)

        for key, value in kwargs.items():
            setattr(self, key, value)





from highton.call_mixins import Call


class CreateCallMixin(Call):
    """
    A mixin to retrieve a single object of a highrise endpoint

    """

    def create(self):
        """
        Creates the object

        :return: the created object
        :rtype: object
        """

        # In the header.location there is the newly created object so get the id out of the url
        created_id = self._post_request(
            data=self.element_to_string(self.encode())
        ).headers.get('Location').split('/')[-1]
        return self.get(created_id)



from abc import ABCMeta

import requests
from requests.auth import HTTPBasicAuth

from highton.highton_constants import HightonConstants
from highton.highton_settings import HightonSettings

class Call(metaclass=ABCMeta):
    ENDPOINT = None
    COLLECTION_DATETIME = '%Y%m%d%H%M%S'

    # yyyymmddhhmmss

    @classmethod
    def _request(cls, method, endpoint=None, params=None, data=None, endpoint_suffix='.xml'):
        response = requests.request(
            method=method,
            url='https://{user}.{highrise_url}/{endpoint}{endpoint_suffix}'.format(
                user=HightonSettings.username(),
                highrise_url=HightonConstants.HIGHRISE_URL,
                endpoint=endpoint if endpoint else cls.ENDPOINT,
                endpoint_suffix=endpoint_suffix
            ),
            headers={'Content-Type': 'application/xml'},
            auth=HTTPBasicAuth(username=HightonSettings.api_key(), password=''),
            params=params,
            data=data,
        )
        response.raise_for_status()
        return response

    @classmethod
    def _get_request(cls, endpoint=None, params=None, endpoint_suffix='.xml'):
        return cls._request(method=HightonConstants.GET, endpoint=endpoint, params=params, endpoint_suffix=endpoint_suffix)

    @classmethod
    def _post_request(cls, data, endpoint=None, params=None):
        return cls._request(method=HightonConstants.POST, endpoint=endpoint, params=params, data=data)

    @classmethod
    def _put_request(cls, data, endpoint=None, params=None):
        return cls._request(method=HightonConstants.PUT, endpoint=endpoint, params=params, data=data)

    @classmethod
    def _delete_request(cls, endpoint=None, params=None):
        return cls._request(method=HightonConstants.DELETE, endpoint=endpoint, params=params)



from highton.models import HightonModel
from highton.highton_constants import HightonConstants
from highton import fields


class ContactData(
    HightonModel
):
    """
    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar phone_numbers: fields.ListField(name=HightonConstants.PHONE_NUMBERS, init_class=PhoneNumber)
    :ivar twitter_accounts: fields.ListField(name=HightonConstants.TWITTER_ACCOUNTS, init_class=TwitterAccount)
    :ivar web_addresses: fields.ListField(name=HightonConstants.WEB_ADDRESSES, init_class=WebAddress)
    :ivar email_addresses: fields.ListField(name=HightonConstants.EMAIL_ADDRESSES, init_class=EmailAddress)
    :ivar addresses: fields.ListField(name=HightonConstants.ADDRESSES, init_class=Address)
    :ivar instant_messenger: fields.ListField(name=HightonConstants.INSTANT_MESSENGERS, init_class=InstantMessenger)

    """
    TAG_NAME = HightonConstants.CONTACT_DATA

    def __init__(self, **kwargs):
        from highton.models import (
            PhoneNumber,
            TwitterAccount,
            WebAddress,
            EmailAddress,
            Address,
            InstantMessenger,
        )

        self.phone_numbers = fields.ListField(name=HightonConstants.PHONE_NUMBERS, init_class=PhoneNumber)
        self.twitter_accounts = fields.ListField(name=HightonConstants.TWITTER_ACCOUNTS, init_class=TwitterAccount)
        self.web_addresses = fields.ListField(name=HightonConstants.WEB_ADDRESSES, init_class=WebAddress)
        self.email_addresses = fields.ListField(name=HightonConstants.EMAIL_ADDRESSES, init_class=EmailAddress)
        self.addresses = fields.ListField(name=HightonConstants.ADDRESSES, init_class=Address)
        self.instant_messenger = fields.ListField(name=HightonConstants.INSTANT_MESSENGERS, init_class=InstantMessenger)

        for key, value in kwargs.items():
            setattr(self, key, value)



from highton.models import HightonModel
from highton.highton_constants import HightonConstants
from highton import fields
from highton import call_mixins


class Comment(
    HightonModel,
    call_mixins.DetailCallMixin,
    call_mixins.CreateCallMixin,
    call_mixins.UpdateCallMixin,
    call_mixins.DeleteCallMixin,
):
    """

    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar parent_id: fields.IntegerField(name=HightonConstants.PARENT_ID)
    :ivar author_id: fields.IntegerField(name=HightonConstants.AUTHOR_ID)
    :ivar created_at: fields.DatetimeField(name=HightonConstants.CREATED_AT)
    :ivar body: fields.StringField(name=HightonConstants.BODY)

    """
    TAG_NAME = HightonConstants.COMMENT
    ENDPOINT = HightonConstants.COMMENTS

    def __init__(self, **kwargs):
        self.parent_id = fields.IntegerField(name=HightonConstants.PARENT_ID, required=True)
        self.body = fields.StringField(name=HightonConstants.BODY, required=True)
        self.author_id = fields.IntegerField(name=HightonConstants.AUTHOR_ID)
        self.created_at = fields.DatetimeField(name=HightonConstants.CREATED_AT)

        super().__init__(**kwargs)



from highton.models import HightonModel
from highton.highton_constants import HightonConstants
from highton import fields


class TwitterAccount(
    HightonModel,
):
    """

    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar location: fields.StringField(name=HightonConstants.LOCATION)
    :ivar username: fields.StringField(name=HightonConstants.USERNAME)
    :ivar url: fields.StringField(name=HightonConstants.URL)
    """
    TAG_NAME = HightonConstants.TWITTER_ACCOUNT

    def __init__(self, **kwargs):
        self.location = fields.StringField(name=HightonConstants.LOCATION)
        self.username = fields.StringField(name=HightonConstants.USERNAME)
        self.url = fields.StringField(name=HightonConstants.URL)

        super().__init__(**kwargs)



from highton.call_mixins import Call
from highton import fields


class ListTaskCallMixin(Call):
    """
    A mixin to get all tasks of inherited class
    These could be: people || companies || kases || deals

    """

    def list_tasks(self):
        """
        Get the tasks of current object

        :return: the tasks
        :rtype: list
        """
        from highton.models.task import Task

        return fields.ListField(
            name=self.ENDPOINT,
            init_class=Task
        ).decode(
            self.element_from_string(
                self._get_request(
                    endpoint=self.ENDPOINT + '/' + str(self.id) + '/' + Task.ENDPOINT,
                ).text
            )
        )



import datetime
from xml.etree import ElementTree

from highton.fields.field import Field
from highton.fields.field_constants import FieldConstants


class DateField(Field):
    """
    The DateField represents a field which value will be parsed to Highrise specific date format

    """
    DATE_FORMAT = '%Y-%m-%d'

    def encode(self):
        """

        :return:
        :rtype: xml.etree.ElementTree.Element
        """
        element = ElementTree.Element(
            self.name,
            attrib={'type': FieldConstants.DATE},
        )
        element = self._set_nil(element, lambda value: value.isoformat())
        return element

    def decode(self, element):
        """

        :param element: the element which needs to be parsed
        :type element: xml.etree.ElementTree.Element
        :return: datetime.date
        :rtype: the parsed date object
        """
        text = element.text
        if text:
            return datetime.datetime.strptime(text, self.DATE_FORMAT).date()
        else:
            return None



class HightonNotLoggedInException(Exception):
    def __str__(self):
        return 'You\'re not logged in, please init HightonSettings'

class HightonSettings:
    @staticmethod
    def return_valid_value_or_raise(value):
        if not value:
            raise HightonNotLoggedInException
        return value

    @classmethod
    def username(cls):
        return HightonSettings.return_valid_value_or_raise(cls.instance.username)

    @classmethod
    def api_key(cls):
        return HightonSettings.return_valid_value_or_raise(cls.instance.api_key)

    class __HightonSettings:
        def __init__(self, username, api_key):
            self.username = username
            self.api_key = api_key

    instance = None

    def __init__(self, username, api_key):
        if not HightonSettings.instance:
            HightonSettings.instance = HightonSettings.__HightonSettings(username, api_key)
        else:
            HightonSettings.instance.username = username
            HightonSettings.instance.api_key = api_key

    def __getattr__(self, name):
        value = getattr(self.instance, name)
        if not value:
            raise HightonNotLoggedInException
        return value


from .highton_model import HightonModel
from .case import Case
from .tag import Tag
from .party import Party
from .contact_data import ContactData
from .address import Address
from .attachment import Attachment
from .phone_number import PhoneNumber
from .twitter_account import TwitterAccount
from .instant_messenger import InstantMessenger
from .email_address import EmailAddress
from .web_address import WebAddress
from .associated_party import AssociatedParty
from .subject_data import SubjectData
from .category import Category
from .deal import Deal
from .task_category import TaskCategory
from .deal_category import DealCategory
from .person import Person
from .company import Company
from .status import Status
from .task import Task
from .user import User
from .group import Group
from .note import Note
from .subject_field import SubjectField



from highton.call_mixins import Call
from highton import fields


class SearchCallMixin(Call):
    """
    A mixin to search models

    """

    SEARCH_OFFSET = 25

    @classmethod
    def search(cls, term=None, page=0, **criteria):
        """
        Search a list of the model
        If you use "term":
        - Returns a collection of people that have a name matching the term passed in through the URL.

        If you use "criteria":
        - returns people who match your search criteria.
        Search by any criteria you can on the Contacts tab, including custom fields. Combine criteria to narrow results

        :param term: params as string
        :type term: str
        :param criteria: search for more criteria
        :type criteria: dict
        :param page: the page
        :type page: int
        :return: the list of the parsed xml objects
        :rtype: list
        """

        assert (term or criteria and not (term and criteria))

        params = {
            'n': int(page) * cls.SEARCH_OFFSET,
        }

        if term:
            params['term'] = term

        if criteria:
            for key, value in criteria.items():
                params['criteria[{}]'.format(key)] = value

        return fields.ListField(name=cls.ENDPOINT, init_class=cls).decode(
            cls.element_from_string(
                cls._get_request(
                    endpoint=cls.ENDPOINT + '/search',
                    params=params
                ).text
            )
        )



class HightonConstants:
    # is used for requests
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    DELETE = 'DELETE'
    HIGHRISE_URL = 'highrisehq.com'

    # Company
    COMPANIES = 'companies'
    COMPANY = 'company'
    COMPANY_NAME = 'company-name'
    COMPANY_ID = 'company-id'

    # Case
    KASES = 'kases'

    # Deal
    DEALS = 'deals'

    TASK = 'task'
    TASKS = 'tasks'

    # Person
    PEOPLE = 'people'
    PERSON = 'person'

    EMAIL = 'email'
    EMAILS = 'emails'

    USER = 'user'
    USERS = 'users'

    GROUP = 'group'
    GROUPS = 'groups'

    NAME = 'name'
    TITLE = 'title'
    FIRST_NAME = 'first-name'
    LAST_NAME = 'last-name'
    CONTACT_DATA = 'contact-data'
    PARENT_ID = 'parent-id'
    EMAIL_ADDRESS = 'email-address'
    EMAIL_ADDRESSES = 'email-addresses'
    TOKEN = 'token'
    DROPBOX = 'dropbox'
    ADMIN = 'admin'

    WEB_ADDRESS = 'web-address'
    WEB_ADDRESSES = 'web-addresses'

    PHONE_NUMBER = 'phone-number'
    PHONE_NUMBERS = 'phone-numbers'

    SUBJECT_DATA = 'subject_data'
    SUBJECT_DATAS = 'subject_datas'

    ADDRESS = 'address'
    ADDRESSES = 'addresses'

    DEAL_CATEGORIES = 'deal_categories'
    DEAL_CATEGORY = 'deal-category'
    TASK_CATEGORIES = 'task_categories'
    TASK_CATEGORY = 'task-category'

    NOTE = 'note'
    NOTES = 'notes'

    TAG = 'tag'
    TAGS = 'tags'

    SUBJECT_ID = 'subject-id'
    SUBJECT_FIELD = 'subject-field'
    SUBJECT_FIELDS = 'subject_fields'
    SUBJECT_FIELD_ID = 'subject_field_id'
    SUBJECT_FIELD_LABEL = 'subject_field_label'

    URL = 'url'
    ZIP = 'zip'
    CITY = 'city'
    STATE = 'state'
    STREET = 'street'
    NUMBER = 'number'
    COUNTRY = 'country'
    LOCATION = 'location'

    ID = 'id'
    BODY = 'body'
    TYPE = 'type'
    VALUE = 'value'
    LABEL = 'label'
    SEARCH = 'search'
    COMMENT = 'comment'
    COMMENTS = 'comments'
    BACKGROUND = 'background'
    RECORDING_ID = 'recording-id'
    FRAME = 'frame'
    ALERT_AT = 'alert-at'
    PUBLIC = 'public'
    RECURRING_PERIOD = 'recurring-period'
    ANCHOR_TYPE = 'anchor-type'
    DONE_AT = 'done-at'

    AUTHOR_ID = 'author-id'
    CLOSED_AT = 'closed-at'
    CREATED_AT = 'created-at'
    UPDATED_AT = 'updated-at'
    VISIBLE_TO = 'visible-to'
    GROUP_ID = 'group-id'
    OWNER_ID = 'owner-id'
    LINKEDIN_URL = 'linkedin-url'
    AVATAR_URL = 'avatar_url'

    ALL = 'all'
    DEAL = 'deal'
    PARTY = 'party'
    PARTIES = 'parties'

    CASE = 'kase'
    CASES = 'kases'

    TWITTER_ACCOUNTS = 'twitter-accounts'
    TWITTER_ACCOUNT = 'twitter-account'

    USERNAME = 'username'
    PROTOCOL = 'protocol'

    COLOR = 'color'

    INSTANT_MESSENGERS = 'instant-messengers'
    INSTANT_MESSENGER = 'instant-messenger'

    ASSOCIATED_PARTIES = 'associated-parties'
    ASSOCIATED_PARTY = 'associated-party'

    ACCOUNT_ID = 'account-id'
    CATEGORY_ID = 'category-id'
    CURRENCY = 'currency'
    DURATION = 'duration'
    PARTY_ID = 'party-id'
    PRICE = 'price'
    PRICE_TYPE = 'price-type'
    RESPONSIBLE_PARTY_ID = 'responsible-party-id'
    STATUS = 'status'
    STATUS_CHANGED_ON = 'status-changed-on'
    CATEGORY = 'category'
    SIZE = 'size'
    DUE_AT = 'due-at'

    ELEMENTS_COUNT = 'elements-count'

    WON = 'won'
    PENDING = 'pending'
    LOST = 'lost'

    COLLECTION_ID = 'collection-id'
    COLLECTION_TYPE = 'collection-type'

    ATTACHMENTS = 'attachments'
    ATTACHMENT = 'attachment'

    SUBJECT_NAME = 'subject-name'
    SUBJECT_TYPE = 'subject-type'
    SUBJECT_TYPES = [COMPANIES, KASES, DEALS, PEOPLE, ]
    CUSTOM_FIELD_TYPES = [PARTY, DEAL, ALL, ]



from highton.call_mixins import Call


class CreateTagCallMixin(Call):
    """
    A mixin to create a tag to inherited class
    These could be: people || companies || kases || deals

    """

    def add_tag(self, name):
        """
        Create a Tag to current object

        :param name: the name of the tag
        :type name: str
        :return: newly created Tag
        :rtype: Tag
        """
        from highton.models.tag import Tag
        created_id = self._post_request(
            endpoint=self.ENDPOINT + '/' + str(self.id) + '/' + Tag.ENDPOINT,
            data=self.element_to_string(
                Tag(name=name).__dict__.get('name').encode()
            )
        ).headers.get('Location').replace('.xml', '').split('/')[-1]
        return Tag.get(created_id)



from highton.call_mixins import Call


class DeleteTagCallMixin(Call):
    """
    A mixin to delete a tag to inherited class
    These could be: people || companies || kases || deals

    """

    def delete_tag(self, tag_id):
        """
        Deletes a Tag to current object

        :param tag_id: the id of the tag which should be deleted
        :type tag_id: int
        :return: newly created Tag
        :rtype: Tag
        """
        from highton.models.tag import Tag
        created_id = self._delete_request(
            endpoint=self.ENDPOINT + '/' + str(self.id) + '/' + Tag.ENDPOINT + '/' + str(tag_id),
        ).headers.get('Location').split('/')[-1]
        return Tag.get(created_id)



from highton import (
    fields,
    call_mixins,
)
from highton.highton_constants import HightonConstants
from highton.models import HightonModel


class Deal(
    HightonModel,
    call_mixins.CreateCallMixin,
    call_mixins.UpdateCallMixin,
    call_mixins.DetailCallMixin,
    call_mixins.ListCallMixin,
    call_mixins.DeleteCallMixin,
    call_mixins.ListNoteCallMixin,
    call_mixins.ListEmailCallMixin,
    call_mixins.ListTagCallMixin,
    call_mixins.CreateTagCallMixin,
    call_mixins.DeleteTagCallMixin,
    call_mixins.ListTaskCallMixin,
    call_mixins.CreateNoteCallMixin,
):
    """
    A deal which represents:

    https://github.com/basecamp/highrise-api/blob/master/sections/deals.md

    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar author_id: fields.IntegerField(name=HightonConstants.AUTHOR_ID)
    :ivar account_id: fields.IntegerField(name=HightonConstants.ACCOUNT_ID)
    :ivar background: fields.StringField(name=HightonConstants.BACKGROUND)
    :ivar category_id: fields.IntegerField(name=HightonConstants.CATEGORY_ID)
    :ivar created_at: fields.DatetimeField(name=HightonConstants.CREATED_AT)
    :ivar currency: fields.StringField(name=HightonConstants.CURRENCY)
    :ivar duration: fields.IntegerField(name=HightonConstants.DURATION)
    :ivar group_id: fields.IntegerField(name=HightonConstants.GROUP_ID)
    :ivar name: fields.StringField(name=HightonConstants.NAME)
    :ivar owner_id: fields.IntegerField(name=HightonConstants.OWNER_ID)
    :ivar party_id: fields.IntegerField(name=HightonConstants.PARTY_ID)
    :ivar price: fields.IntegerField(name=HightonConstants.PRICE)
    :ivar price_type: fields.StringField(name=HightonConstants.PRICE_TYPE)
    :ivar responsible_party_id: fields.IntegerField(name=HightonConstants.RESPONSIBLE_PARTY_ID)
    :ivar status: fields.StringField(name=HightonConstants.STATUS)
    :ivar status_changed_on: fields.DateField(name=HightonConstants.STATUS_CHANGED_ON)
    :ivar updated_at: fields.DatetimeField(name=HightonConstants.UPDATED_AT)
    :ivar visible_to: fields.StringField(name=HightonConstants.VISIBLE_TO)
    :ivar party: fields.ObjectField(name=HightonConstants.PARTY, init_class=Party)
    :ivar category: fields.ObjectField(name=HightonConstants.CATEGORY, init_class=Category)
    :ivar tags: fields.ListField(name=HightonConstants.TAGS, init_class=Tag)
    :ivar parties: fields.ListField(name=HightonConstants.PARTIES, init_class=Party)
    :ivar contact_data: fields.ObjectField(name=HightonConstants.CONTACT_DATA, init_class=ContactData)
    :ivar subject_datas: fields.ListField(name=HightonConstants.SUBJECT_DATAS, init_class=SubjectData)
    :ivar associated_parties: fields.ListField(name=HightonConstants.ASSOCIATED_PARTIES, init_class=AssociatedParty)
    """

    ENDPOINT = HightonConstants.DEALS
    TAG_NAME = HightonConstants.DEAL
    OFFSET = 500

    def __init__(self, **kwargs):
        """
        Init a deal with the attributes

        :param kwargs:
        :type kwargs:
        """
        from highton.models import (
            Tag,
            ContactData,
            SubjectData,
            Party,
            AssociatedParty,
            Category,
        )

        self.author_id = fields.IntegerField(name=HightonConstants.AUTHOR_ID)
        self.account_id = fields.IntegerField(name=HightonConstants.ACCOUNT_ID)
        self.background = fields.StringField(name=HightonConstants.BACKGROUND)
        self.category_id = fields.IntegerField(name=HightonConstants.CATEGORY_ID)
        self.created_at = fields.DatetimeField(name=HightonConstants.CREATED_AT)
        self.currency = fields.StringField(name=HightonConstants.CURRENCY)
        self.duration = fields.IntegerField(name=HightonConstants.DURATION)
        self.group_id = fields.IntegerField(name=HightonConstants.GROUP_ID)
        self.name = fields.StringField(name=HightonConstants.NAME)
        self.owner_id = fields.IntegerField(name=HightonConstants.OWNER_ID)
        self.party_id = fields.IntegerField(name=HightonConstants.PARTY_ID)
        self.price = fields.IntegerField(name=HightonConstants.PRICE)
        self.price_type = fields.StringField(name=HightonConstants.PRICE_TYPE)
        self.responsible_party_id = fields.IntegerField(name=HightonConstants.RESPONSIBLE_PARTY_ID)
        self.status = fields.StringField(name=HightonConstants.STATUS)
        self.status_changed_on = fields.DateField(name=HightonConstants.STATUS_CHANGED_ON)
        self.updated_at = fields.DatetimeField(name=HightonConstants.UPDATED_AT)
        self.visible_to = fields.StringField(name=HightonConstants.VISIBLE_TO)
        self.party = fields.ObjectField(name=HightonConstants.PARTY, init_class=Party)
        self.category = fields.ObjectField(name=HightonConstants.CATEGORY, init_class=Category)
        self.tags = fields.ListField(name=HightonConstants.TAGS, init_class=Tag)
        self.parties = fields.ListField(name=HightonConstants.PARTIES, init_class=Party)
        self.contact_data = fields.ObjectField(name=HightonConstants.CONTACT_DATA, init_class=ContactData)
        self.subject_datas = fields.ListField(name=HightonConstants.SUBJECT_DATAS, init_class=SubjectData)
        self.associated_parties = fields.ListField(name=HightonConstants.ASSOCIATED_PARTIES, init_class=AssociatedParty)

        super().__init__(**kwargs)

    @classmethod
    def list(cls, page=0, status=None, since=None):
        """

        :param page: page starting by 0
        :type page: int
        :param since:
        :type since: datetime.datetime
        :param status:
        :type status: str
        :return: list of person objects
        :rtype: list
        """
        params = {}
        if page:
            params['n'] = int(page) * cls.OFFSET
        if since:
            params['since'] = since.strftime(cls.COLLECTION_DATETIME)
        if status:
            params['status'] = status

        return super().list(params)

    def update_status(self, status):
        """
        Updates the status of the deal

        :param status: status have to be ('won', 'pending', 'lost')
        :return: successfull response or raise Exception
        :rtype:
        """
        assert (status in (HightonConstants.WON, HightonConstants.PENDING, HightonConstants.LOST))
        from highton.models import Status

        return self._put_request(
            data=Status(name=status),
            endpoint=self.ENDPOINT + '/' + str(self.id) + '/status',
        )



from xml.etree import ElementTree

from highton.fields import Field
from highton.fields.field_constants import FieldConstants


class ListField(Field):
    """
    The ListField parses the init_class objects in a list

    :ivar init_class: a Highton model
    """

    def __init__(self, name, init_class):
        super().__init__(name)
        self.init_class = init_class

    def encode(self):
        """
        Just iterate over the child elements and append them to the current element

        :return: the encoded element
        :rtype: xml.etree.ElementTree.Element
        """
        element = ElementTree.Element(
            self.name,
            attrib={'type': FieldConstants.ARRAY},
        )
        for item in self.value:
            element.append(item.encode())
        return element

    def decode(self, element):
        """

        :param element:
        :type element:
        :return: the parsed list with init_class objects
        :rtype: list
        """
        child_elements = []
        for child in element:
            child_elements.append(self.init_class.decode(child))
        return child_elements







from highton import call_mixins
from highton import fields
from highton.highton_constants import HightonConstants
from highton.models import HightonModel


class Tag(
    HightonModel,
    call_mixins.ListCallMixin,
    call_mixins.DetailCallMixin,
    call_mixins.DeleteCallMixin,
):
    """
    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar name: fields.StringField(name=HightonConstants.NAME)
    """
    ENDPOINT = HightonConstants.TAGS
    TAG_NAME = HightonConstants.TAG

    def __init__(self, **kwargs):
        self.name = fields.StringField(name=HightonConstants.NAME)

        super().__init__(**kwargs)

    @classmethod
    def get(cls, object_id):
        """
        Get all parties (people and companies) associated with a given tag.
        :param object_id: the primary id of the model
        :type object_id: integer
        :return: the parties
        :rtype: list
        """
        from highton.models.party import Party
        return fields.ListField(name=Party.ENDPOINT, init_class=Party).decode(
            cls.element_from_string(
                cls._get_request(endpoint=cls.ENDPOINT + '/' + str(object_id)).text
            )
        )



from highton import (
    fields,
    call_mixins,
)
from highton.highton_constants import HightonConstants
from highton.models import HightonModel


class TypeCategory(
    HightonModel,
    call_mixins.ListCallMixin,
    call_mixins.DetailCallMixin,
    call_mixins.CreateCallMixin,
    call_mixins.UpdateCallMixin,
    call_mixins.DeleteCallMixin,
):
    """

    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar name: fields.StringField(name=HightonConstants.NAME)
    :ivar color: fields.StringField(name=HightonConstants.COLOR)
    :ivar account_id: fields.IntegerField(name=HightonConstants.ACCOUNT_ID)
    :ivar created_at: fields.DatetimeField(name=HightonConstants.CREATED_AT)
    :ivar updated_at: fields.DatetimeField(name=HightonConstants.UPDATED_AT)
    :ivar elements_count: fields.IntegerField(name=HightonConstants.ELEMENTS_COUNT)
    """
    def __init__(self, **kwargs):
        self.name = fields.StringField(name=HightonConstants.NAME, required=True)
        self.color = fields.StringField(name=HightonConstants.COLOR)
        self.account_id = fields.IntegerField(name=HightonConstants.ACCOUNT_ID)
        self.created_at = fields.DatetimeField(name=HightonConstants.CREATED_AT)
        self.updated_at = fields.DatetimeField(name=HightonConstants.UPDATED_AT)
        self.elements_count = fields.IntegerField(name=HightonConstants.ELEMENTS_COUNT)
        super().__init__(**kwargs)



from abc import ABCMeta
from xml.etree import ElementTree

from highton import fields


class XMLEncoder(metaclass=ABCMeta):

    """
    Encodes the object to an xml.etree.ElementTree.Element

    """

    TAG_NAME = None

    def __setattr__(self, key, value):
        if key in self.__dict__ and isinstance(self.__dict__[key], fields.Field):
            self.__dict__[key].value = value
        else:
            object.__setattr__(self, key, value)

    @staticmethod
    def element_to_string(element):
        """
        Parses the xml.etree.ElementTree.Element to a string

        :param element: a xml element
        :type element: xml.etree.ElementTree.Element
        :return: the parsed string
        :rtype: str
        """
        return ElementTree.tostring(element)

    def encode(self):
        """
        Encodes the object to a xml.etree.ElementTree.Element

        :return: the encoded element
        :rtype: xml.etree.ElementTree.Element
        """
        root_element = ElementTree.Element(self.TAG_NAME)
        for value in [value for value in self.__dict__.values() if isinstance(value, fields.Field)]:
            if value.required or value.value:
                root_element.append(value.encode())
        return root_element



from highton.fields import Field


class ObjectField(Field):
    """
    The ObjectFIeld parses the init_class object

    :ivar init_class: a Highton model
    """
    def __init__(self, name, init_class):
        super().__init__(name)
        self.init_class = init_class

    def encode(self):
        """
        Just encode the object you have as value
        
        :return: the parsed element
        :rtype: xml.etree.ElementTree.Element
        """
        return self.value.encode()

    def decode(self, element):
        return self.init_class.decode(element)



from highton.models import Party
from highton.highton_constants import HightonConstants


class AssociatedParty(
    Party,
):
    """

    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar author_id: fields.IntegerField(name=HightonConstants.AUTHOR_ID)
    :ivar background: fields.StringField(name=HightonConstants.BACKGROUND)
    :ivar company_id: fields.IntegerField(name=HightonConstants.COMPANY_ID)
    :ivar created_at: fields.DatetimeField(name=HightonConstants.CREATED_AT)
    :ivar first_name: fields.StringField(name=HightonConstants.FIRST_NAME)
    :ivar name: fields.StringField(name=HightonConstants.NAME)
    :ivar group_id: fields.IntegerField(name=HightonConstants.GROUP_ID)
    :ivar last_name: fields.StringField(name=HightonConstants.LAST_NAME)
    :ivar owner_id: fields.IntegerField(name=HightonConstants.OWNER_ID)
    :ivar title: fields.StringField(name=HightonConstants.TITLE)
    :ivar updated_at: fields.DatetimeField(name=HightonConstants.UPDATED_AT)
    :ivar visible_to: fields.StringField(name=HightonConstants.VISIBLE_TO)
    :ivar company_name: fields.StringField(name=HightonConstants.COMPANY_NAME)
    :ivar linkedin_url: fields.StringField(name=HightonConstants.LINKEDIN_URL)
    :ivar avatar_url: fields.StringField(name=HightonConstants.AVATAR_URL)
    :ivar type: fields.StringField(name=HightonConstants.TYPE)
    :ivar tags: fields.ListField(name=HightonConstants.TAGS, init_class=Tag)
    :ivar contact_data: fields.ObjectField(name=HightonConstants.CONTACT_DATA, init_class=ContactData)
    :ivar subject_datas: fields.ListField(name=HightonConstants.SUBJECT_DATAS, init_class=SubjectData)
    """
    TAG_NAME = HightonConstants.ASSOCIATED_PARTY



from highton import call_mixins
from highton.models import HightonModel
from highton.highton_constants import HightonConstants
from highton import fields


class Group(
    HightonModel,
    call_mixins.ListCallMixin,
    call_mixins.DetailCallMixin,
    call_mixins.CreateCallMixin,
    call_mixins.UpdateCallMixin,
    call_mixins.DeleteCallMixin,
):
    """
    The authenticated user needs to be an administrator to perform these actions.

    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar name: fields.StringField(name=HightonConstants.NAME)
    :ivar users: fields.ListField(name=HightonConstants.USERS, init_class=models.User)
    """
    ENDPOINT = HightonConstants.GROUPS
    TAG_NAME = HightonConstants.GROUP

    def __init__(self, **kwargs):
        from highton import models

        self.name = fields.StringField(name=HightonConstants.NAME)
        self.users = fields.ListField(name=HightonConstants.USERS, init_class=models.User)

        super().__init__(**kwargs)



from highton import fields
from highton.models import HightonModel
from highton.highton_constants import HightonConstants


class Contact(
    HightonModel,

):
    """
    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar author_id: fields.IntegerField(name=HightonConstants.AUTHOR_ID)
    :ivar background: fields.StringField(name=HightonConstants.BACKGROUND)
    :ivar created_at: fields.DatetimeField(name=HightonConstants.CREATED_AT)
    :ivar group_id: fields.IntegerField(name=HightonConstants.GROUP_ID)
    :ivar owner_id: fields.IntegerField(name=HightonConstants.OWNER_ID)
    :ivar updated_at: fields.DatetimeField(name=HightonConstants.UPDATED_AT)
    :ivar visible_to: fields.StringField(name=HightonConstants.VISIBLE_TO)
    :ivar linkedin_url: fields.StringField(name=HightonConstants.LINKEDIN_URL)
    :ivar avatar_url: fields.StringField(name=HightonConstants.AVATAR_URL)
    :ivar tags: fields.ListField(name=HightonConstants.TAGS, init_class=Tag)
    :ivar contact_data: fields.ObjectField(name=HightonConstants.CONTACT_DATA, init_class=ContactData)
    :ivar subject_datas: fields.ListField(name=HightonConstants.SUBJECT_DATAS, init_class=SubjectData)
    """
    def __init__(self, **kwargs):
        from highton.models import (
            Tag,
            ContactData,
            SubjectData,
        )

        self.author_id = fields.IntegerField(name=HightonConstants.AUTHOR_ID)
        self.background = fields.StringField(name=HightonConstants.BACKGROUND)
        self.created_at = fields.DatetimeField(name=HightonConstants.CREATED_AT)
        self.group_id = fields.IntegerField(name=HightonConstants.GROUP_ID)
        self.owner_id = fields.IntegerField(name=HightonConstants.OWNER_ID)
        self.updated_at = fields.DatetimeField(name=HightonConstants.UPDATED_AT)
        self.visible_to = fields.StringField(name=HightonConstants.VISIBLE_TO)
        self.linkedin_url = fields.StringField(name=HightonConstants.LINKEDIN_URL)
        self.avatar_url = fields.StringField(name=HightonConstants.AVATAR_URL)
        self.tags = fields.ListField(name=HightonConstants.TAGS, init_class=Tag)
        self.contact_data = fields.ObjectField(name=HightonConstants.CONTACT_DATA, init_class=ContactData)
        self.subject_datas = fields.ListField(name=HightonConstants.SUBJECT_DATAS, init_class=SubjectData)

        super().__init__(**kwargs)



from highton.call_mixins import Call
from highton import fields


class DetailCallMixin(Call):
    """
    A mixin to retrieve a single object of a highrise endpoint

    """

    @classmethod
    def get(cls, object_id):
        """
        Retrieves a single model

        :param object_id: the primary id of the model
        :type object_id: integer
        :return: the object of the parsed xml object
        :rtype: object
        """
        return fields.ObjectField(name=cls.ENDPOINT, init_class=cls).decode(
            cls.element_from_string(
                cls._get_request(endpoint=cls.ENDPOINT + '/' + str(object_id)).text
            )
        )



from highton.call_mixins import Call
from highton import fields


class ListCallMixin(Call):
    """
    A mixin to retrieve a list of a highrise endpoint

    """

    @classmethod
    def list(cls, params=None):
        """
        Retrieves a list of the model

        :param params: params as dictionary
        :type params: dict
        :return: the list of the parsed xml objects
        :rtype: list
        """
        return fields.ListField(name=cls.ENDPOINT, init_class=cls).decode(
            cls.element_from_string(cls._get_request(params=params).text)
        )



from highton.call_mixins import Call


class DeleteCallMixin(Call):
    """
    A mixin to delete a single object of a highrise endpoint

    """

    def delete(self):
        """
        Deletes the object

        :return:
        :rtype: None
        """
        return self._delete_request(endpoint=self.ENDPOINT + '/' + str(self.id))



from highton.highton_constants import HightonConstants
from highton.models.type_category import TypeCategory


class DealCategory(
    TypeCategory,
):
    """
    :ivar id: fields.IntegerField(name=HightonConstants.ID)
    :ivar name: fields.StringField(name=HightonConstants.NAME)
    :ivar color: fields.StringField(name=HightonConstants.COLOR)
    :ivar account_id: fields.IntegerField(name=HightonConstants.ACCOUNT_ID)
    :ivar created_at: fields.DatetimeField(name=HightonConstants.CREATED_AT)
    :ivar updated_at: fields.DatetimeField(name=HightonConstants.UPDATED_AT)
    :ivar elements_count: fields.IntegerField(name=HightonConstants.ELEMENTS_COUNT)
    """
    ENDPOINT = HightonConstants.DEAL_CATEGORIES
    TAG_NAME = HightonConstants.DEAL_CATEGORY
